#!/usr/bin/lua
-- Contains all functions that processor uses

----------------------------------------------
-- Check if machine is a poller
-- @return bool
----------------------------------------------
function is_poller()
	return (os.execute("(/usr/bin/mon node show --type=master|/bin/grep -q 'TYPE=master') 2> /dev/null", 'r') == 0)
end

----------------------------------------------
-- Get a random number
-- @param vararg
-- @return int
----------------------------------------------
function real_random(...)
	return math.random(...)
end

----------------------------------------------
-- Sleep for supplied seconds
-- @param sec int
-- @return void
----------------------------------------------
function sleep(sec)
	socket.select(nil, nil, sec)
end

----------------------------------------------
-- Remove a file or directory recursively (not removing special files)
-- Does not trigger error(), you only need to rely on return value(s)
-- @param string file- or dirname
-- @return boolean, string|nil (error message)
----------------------------------------------
function rm_rf(to_delete)
	local attrs = lfs.attributes(to_delete)
	if attrs == nil then
		return true
	end
	if attrs.mode == "file" then
		if not os.remove(to_delete) then
			return false, string.format("Failed to remove '%s' (check permissions)", to_delete)
		end
		return true
	end
	if attrs.mode == "directory" then
		local file
		for file in lfs.dir(to_delete) do
			if file ~= "." and file ~= ".." then
				local status, msg
				--print(string.format("Triggering recursively '%s'", to_delete))
				status, msg = rm_rf(string.format("%s/%s", to_delete, file))
				if not status then
					return status, msg
				end
			end
		end
		if not lfs.rmdir(to_delete) then
			return false, string.format("Failed to remove directory '%s' (check permissions)", to_delete)
		end
		return true
	end
	return false, string.format("Not removing '%s' '%s', need a file or directory", attrs.mode, to_delete)
end

----------------------------------------------
-- Documented function
-- @param res string
-- @param inherits metatable function
-- @return table
----------------------------------------------
function response_to_obj(res, inherits, ...)
	local s = socket.gettime()
	if type(res) ~= "string" then
		SYSError("Failed unserializing livestatus argument")
		return nil
	end
	local status, res, msg = pcall(json.decode, res)
	if not status then
		SYSError("Failed decoding livestatus json")
		return nil
	end
	local inst = {}
	if #res < 2 then
		return nil, "Requested object not found"
	end
	if inherits then
		setmetatable(inst, { __index = inherits })
	end
	for i, v in pairs(res[1]) do
		inst[v] = res[2][i]
	end
	return inst
end

----------------------------------------------
-- Log to syslog if config.debug is true
-- @param msg string
-- @return void
----------------------------------------------
function SYSDebug(msg, ...)
	if not config.debug then
		return
	end
	log.syslog("LOG_NOTICE", msg, ...)
	print(os.time(), "DEBUG ", msg, ...)
end

----------------------------------------------
-- Log errors to syslog and stderr
-- @param msg string
-- @return void
----------------------------------------------
function SYSError(msg, ...)
	log.syslog("LOG_ERR", msg, ...)
	io.stderr:write(msg)
end

----------------------------------------------
-- Clear table of cached states
-- @return void
----------------------------------------------
function clear_state_table()
	SYSDebug("Clearing state table.")
	states = {}
end

----------------------------------------------
-- Gets configurable livestatus socket path
-- @return string
----------------------------------------------
function get_socket_path()
	local socket_path = nil
	local file = io.open(config.livestatus_config_path, "r")
	if file then
		-- Get configured socket path
		for line in file:lines() do
			if string.sub(line, 0, 5) == "path:" then
				socket_path = string.match(line, "[^%s]+/.*")
			end
		end
		file:close()
	end
	-- Default
	if not socket_path then
		socket_path = config.default_ls_path
	end
	return socket_path
end

----------------------------------------------
-- Process all changes to trees
-- @return void
----------------------------------------------
function process_rules_spool_all()
	if lfs.attributes(config.spool_dir).mode ~= "directory" then
		error("The spool_dir config option ("..config.spool_dir..") must be a directory")
	end
	for uuid in lfs.dir(config.spool_dir) do
		if (uuid ~= "." and uuid ~= "..") then
			if (lfs.attributes(config.spool_dir..uuid).mode == "directory") then
				process_rules_spool(uuid)
				local status, msg = rm_rf(config.spool_dir..uuid)
				if not status then
					SYSError(msg)
				end
			end
		end
	end
	SYSDebug("Finished checking for spool files")
end

----------------------------------------------
-- Process changes to trees in specific folder
-- @param uuid string
-- @return void
----------------------------------------------
function process_rules_spool(uuid)
	local c = 0
	local spool_dir = config.spool_dir..uuid
	for file in lfs.dir(spool_dir) do
		if (file ~= "." and file ~= ".." and file:sub(-4) == ".lua") then
			local s, m = os.execute("mv -f \""..spool_dir.."/"..file.."\" "..config.tree_dir)
			if s then
				c = c + 1
				SYSDebug("New rule "..file:sub(0, -5).." successfully loaded.")
			else
				SYSError("Cannot move new "..file:sub(0, -5).." rule in working dir: "..m)
			end
		end
	end
	SYSDebug("Spool folder '"..spool_dir.."' contained '"..c.."' new top level elements.")
	local del_list = io.open(spool_dir.."/delete.list")
	if not del_list then
		SYSDebug("No delete.list file in spool dir '"..spool_dir.."'")
		return
	end
	local s, r = pcall(del_list.read, del_list, "*all")
	if not s then
		SYSError("Spool delete list content reading error. List ignored.")
		return
	end
	s, r = pcall(json.decode, r)
	if not s then
		SYSError("Spool delete list json decoding error. List ignored.")
		return
	end
	for i, v in ipairs(r) do
		if type(v) == "string" then
			local status, msg = rm_rf(config.tree_dir..v..".lua")
			if status then
				SYSDebug("Rule "..v.." successfully deleted.")
			else
				SYSError("Deleting rule "..v.." error: "..msg)
			end
		else
			SYSError("Non string value in delete list with index #"..i.." (found '"..type(v).."')")
		end
	end
end

----------------------------------------------
-- Executes rules
-- @param rule string
-- @param output bool (true: debug mode, false = execute passive check if possible)
-- @return table
----------------------------------------------
function execute_rule(rule, output)
	local passive = true
	local output_file = io.stdout
	local result = {};
	local result_top = {};
	result.items = {}
	result_top.items = {}

	if (rule == "all") then
		status, res, msg = pcall(get_rule_list)
	else
		res = { rule }
		status = true
	end
	if output then
		passive = false
	end
	if status then
		for i, rule in ipairs(res) do
			local time_mem = socket.gettime()
			local loop_start = time_mem

			-- this rule loadfile must MUST set the variable "script"
			local _, res, msg = pcall(loadfile, config.tree_dir..rule..".lua")

			SYSDebug(string.format("Load rule complete. %.4f sec", socket.gettime() - time_mem))
			time_mem = socket.gettime()
			if res == nil then
				SYSError("Could not load rule file for rule "..rule..": ".. msg)
				error()
			else
				-- actually the variable "script" is set here when executing the returned function of loadfile
				res()
				SYSDebug(string.format("Execution of rule complete. %.4f sec", socket.gettime() - time_mem))
			end
			time_mem = socket.gettime()

			local exec_s, r, m = execute(script, "", passive)
			if exec_s then
				SYSDebug(string.format("Rule processing complete. %.4f sec", socket.gettime() - time_mem))
			else
				SYSError("Execution of rule failed: "..m)
			end

			-- Calculate full tree results
			local prep_s, res, msg = pcall(prepare_full, script)
			if not prep_s then
				SYSError("Json prepare failed: "..(res or msg) or "")
				error()
			end
			SYSDebug("Full result for rule "..rule.." generated")
			if res then
				result.items[rule] = res
			end

			-- Generate top level results for widget
			local status, topresult, msg = pcall(prepare_top, script)
			if status then
				result_top.items[rule] = topresult
				SYSDebug("Top result for rule "..rule.." generated")
			else
				SYSDebug("Prepare top json for rule "..rule.." failed")
			end

			SYSDebug(string.format("Rule executed in %.4f sec", socket.gettime() - loop_start))
		end
	else
		SYSError("Failed getting rule list: "..(res or msg) or "")
	end

	cached_full = json.encode(result)
	cached_top = json.encode(result_top)

	return r
end

function exists_in_livestatus(host_name, service_description)
	h = host(host_name)
	if not h then
		return false
	end

	s = h:service(service_description)
	if not s then
		return false
	end
	return true
end

----------------------------------------------
-- Format calculated result to full tree
-- @param obj table
-- @return table
----------------------------------------------
function prepare_full(obj)
	l = l or 0
	local result = {}
	local c = 0
	result.name = obj["name"]
	result.type = obj["intent"]["type"]
	result.rule = obj["intent"]["name"]
	result.args = {}
	for k,v in ipairs(obj["intent"]["args"] or {}) do
		table.insert(result.args, v)
	end
	result["materialized"] = obj["materialized"] or nil
	-- bubble "'materialized' according to synergy config, but non-existing
	-- according to livestatus", to the GUI
	result["materialized_and_exists_in_livestatus"] = false
	if obj["materialized"] then
		result["materialized_and_exists_in_livestatus"] = exists_in_livestatus(obj["materialized"]["host"], obj["name"])
	end
	if type(obj["components"]) == "table" then
		result.items = {}
		for k,v in ipairs(obj["components"]) do
			local itemname = v.name
			result.items[itemname] = prepare_full(v)
		end
	end
	if type(obj["result"]) ~= "table" or table.getn(obj["result"]) == 0 then
		result.status = "UNKNOWN"
		result.result = {}
		result.result.msg = "No result"
		result.result.status = 3
	else
		result.status = status_map[obj["result"][1]["status"] + 1]
		result.result = {}
		for k,v in pairs(obj["result"][1]) do
			result.result[k] = v
		end
	end
	return result
end

----------------------------------------------
-- Format calculated top level result
-- @param obj table
-- @return table
----------------------------------------------
function prepare_top(obj)
	SYSDebug("Preparing top json "..obj["name"])
	local res = {}
	res["name"] = obj["name"]
	res["type"] = obj["intent"]["type"]
	res["rule"] = obj["intent"]["name"]
	res["args"] = obj["intent"]["args"]
	res["materialized"] = obj["materialized"]
	-- bubble "'materialized' according to synergy config, but non-existing
	-- according to livestatus", to the GUI
	res["materialized_and_exists_in_livestatus"] = false
	if obj["materialized"] then
		res["materialized_and_exists_in_livestatus"] = exists_in_livestatus(obj["materialized"]["host"], obj["name"])
	end
	if type(obj["result"]) ~= "table" or table.getn(obj["result"]) == 0 then
		res["result"] = {status = "UNKNOWN", result = {msg = "no result", status = 3}}
	else
		res["result"] = obj["result"][1]
	end
	return res
end

----------------------------------------------
-- Strips full result to necessary data for reports
-- @param data table
-- @return table
-- @TODO Obviously this is not the way to do this.
-- Fomrmat the data from the object instead.
----------------------------------------------
function parse_report_data(data)
	if type(data) ~= "table" or type(data.items) ~= "table" then
		-- Tree might be empty, return false.
		return false
	end
	for k,v in pairs(data.items) do
		if v.items then
			v.status = v.result.status
			v.result.status = nil
			v.type = nil
			v.args = nil
			v.result.last_check = nil
			v.rule = nil
			data.items[k] = parse_report_data(v)
		elseif v.status == "OK" then
			-- Remove Ok objects
			data.items[k] = nil
		else
			-- Strip data that's not needed
			v.rule = nil
			v.type = nil
			v.status = v.result.status
			v.result = nil
			v.args = nil
		end
	end
	return data
end

----------------------------------------------
-- Write passive check to nagios command pipe
-- @param cmd string
-- @return void
----------------------------------------------
function passive_check_proccess(cmd)
	local pipe, msg = io.open(config.nagios_cmd_path, "a+")
	pipe:write(cmd)
	pipe:close()
end

----------------------------------------------
-- Format passive check result
-- @param host_name string
-- @param service_dsc string
-- @param code int
-- @param msg string
-- @return void
----------------------------------------------
function passive_check(timestamp, host_name, service_dsc, code, msg)
	local cmd = string.format("[%d] PROCESS_SERVICE_CHECK_RESULT;%s;%s;%d;%s;\n", timestamp, host_name, service_dsc, code, msg)
	local status, res, msg
	local tries = 0
	while tries < config.max_retries do
		status, res, msg = pcall(passive_check_proccess, cmd)
		if status then
			return
		end
		tries = tries + 1
		SYSError("Passive check: "..cmd.." failed. Error: "..(msg or ""))
		SYSDebug("Trying again. "..tries.."/"..config.max_retries)
		sleep(config.wait_on_fail)
	end
end

----------------------------------------------
-- Fetch all trees as tables
-- @return table
----------------------------------------------
function get_rule_list()
	local res = {}
	for file in lfs.dir(config.tree_dir) do
		if (file ~= "." and file ~= ".." and file:sub(-4) == ".lua") then
			table.insert(res, file:sub(0, -5))
		end
	end
	return res
end

----------------------------------------------
-- Intersect contacts from contactgroups
-- @param list table
-- @return string
----------------------------------------------
function intersection(list)
	if #list == 0 then
		return {}
	end
	local list_count = 0
	local items = {}
	for k,v in ipairs(list) do
		if #v == 0 then
			return {}
		end
		if (v ~= "*") and (not ((#v == 1) and (v[1] == "*"))) then
			list_count = list_count + 1
			local used = {}
			for sk, sv in ipairs(v) do
				if not used[sv] then
					table.insert(items, sv)
					used[sv] = true
				end
			end
		end
	end
	if #items == 0 then
		return "*"
	end
	table.sort(items)
	local last_v
	local last_l = 0
	local res = {}
	for k,v in ipairs(items) do
		if last_v == v then
			last_l = last_l + 1
		else
			if last_l == list_count then
				table.insert(res, last_v)
			end
			last_l = 1
			last_v = v
		end
	end
	if last_l == list_count then
		table.insert(res, last_v)
	end
	return res
end

----------------------------------------------
-- Entry point for executing rules on objects
--
-- @param obj table
-- @param path string
-- @param passive bool
-- @return table|nil
----------------------------------------------
function execute(obj, path, passive)
	local new_path = path.."/"..obj.name
	local arguments = {}
	if type(obj.components) == "table" then
		for i,v in ipairs(obj.components) do
			-- Process each child node, depth first
			local res = execute(v, new_path, passive)
			if type(res) == "table" then
				arguments = table_merge(arguments, res)
			end
		end
	end

	if type(obj.intent) == "table" then
		return run_rule(obj, new_path, arguments, passive)
	end
end

----------------------------------------------
-- Entry point for executing rules on objects
--
-- @param obj table
-- @param path string
-- @param passive bool
-- @return table
----------------------------------------------
function run_rule(obj, new_path, arguments, passive)
	local result
	local status, msg
	local timestamp = os.time()
	obj.result = obj.intent.func(arguments, unpack(obj.intent.args))
	result = obj.result[1]
	-- TODO remove all traces of "items"? seems to be unused
	items = {}

	if (obj.intent.type ~= "nagios") then
		set_last_state_change(timestamp, states, new_path, result)
	end

	if obj.materialized then
		try_to_send_passive_check_result(obj, passive, result, items, timestamp)
	end

	-- Object either a) not materialized, or b) have no contacts on them
	-- The rest of our code wants contacts
	if not result.host_contacts then
		result.host_contacts = {}
	end
	if not result.service_contacts then
		result.service_contacts = {}
	end

	return obj.result
end

function set_last_state_change(timestamp, states, new_path, result)
	-- lsc => last state change
	if type(states[new_path]) == "table" then
		if states[new_path].status ~= result.status then
			states[new_path].status = result.status
			result.last_state_change = timestamp
		else
			result.last_state_change = states[new_path].lsc
		end
	else
		states[new_path] = {}
		states[new_path].status = result.status
		states[new_path].lsc = timestamp
		result.last_state_change = states[new_path].lsc
	end
	result.last_check = timestamp
end

function try_to_send_passive_check_result(obj, passive, result, items, timestamp)
	-- Object is materialized, fetch last state change from
	-- the materialized service.
	--
	-- TODO host() should be injected or mocked in any other way
	mhost = host(obj.materialized.host)
	if not mhost then
		SYSError("Could not find host for materialized service. Ignoring materialization of "..(obj.materialized.host or "N/A"))
		return
	end

	-- TODO service() should be injected or mocked in any other way
	mserv = mhost:service(obj.name)
	if not mserv then
		SYSError("Materialized service does not exist: "..obj.materialized.host..";"..obj.name)
		pcall(materialize_service, obj.materialized.host, obj.name)
		return
	end
	local mserv_contacts = mserv.contacts
	local mserv_contact_groups = mserv.contact_groups
	-- Set object contacts
	local s, m
	local materialized_host_contacts = {}
	local mhost_contacts = mhost.contacts
	local mhost_contact_groups = mhost.contact_groups
	mhost_contacts = table_merge(mhost_contacts, contactgroups_get_contacts(mhost_contact_groups))
	table.insert(materialized_host_contacts, mhost_contacts)
	s, obj.result[1].host_contacts, m = pcall(intersection, materialized_host_contacts)
	local materialized_service_contacts = {}
	mserv_contacts = table_merge(mserv_contacts, contactgroups_get_contacts(mserv_contact_groups))
	table.insert(materialized_service_contacts, mserv_contacts)
	s, obj.result[1].service_contacts, m = pcall(intersection, materialized_service_contacts)
	if passive then
		passive_check(os.time(), obj.materialized.host, obj.name, result.status, result.msg)
		local res = calculate_full_tree(obj, items)
		save_report_data(mserv, result, obj, res, timestamp, write_report_data_to_php_controller)
	end
end

function calculate_full_tree(obj, items)
	-- Calculate full tree results
	local prep_s, res, msg = pcall(prepare_full, obj)
	if not prep_s then
		SYSError("Json prepare failed: "..(res or msg) or "")
		error()
	end

	SYSDebug("Full result for rule "..obj.name.." generated")

	if res then
		-- items never used?
		items[obj.name] = res
	end
	return res
end


----------------------------------------------
-- Should the result trigger a report event? If so, execute write_callback()
--
-- @param old_service_data table
-- @param result table
-- @param obj table
-- @param res table
-- @param timetamp number
-- @param write_callback function|nil
-- @return boolean
----------------------------------------------
function save_report_data(old_service_data, result, obj, res, timestamp, write_callback)
	if old_service_data == nil or next(old_service_data) == nil then
		SYSError("save_report_data() needs status from a service, it might have been remove or renamed: "..obj.materialized.host..";"..obj.name)
		return false
	end

	if not config.save_report_data then
		SYSDebug("Not saving report data: "..json.encode({
			name = obj.name,
			config = config.save_report_data,
			mserv_checked = old_service_data.has_been_checked,
			mserv_state = old_service_data.state,
			result_status = result.status
		}))
		return false
	end

	-- Always log the first check of a service (has_been_checked == 0), so
	-- ignore the state in that case
	if old_service_data.has_been_checked == 1 and old_service_data.state == result.status then
		SYSDebug("Not saving report data: "..json.encode({
			name = obj.name,
			config = config.save_report_data,
			mserv_checked = old_service_data.has_been_checked,
			mserv_state = old_service_data.state,
			result_status = result.status
		}))
		return false
	end
	local parsed_data = { data = "" }
	if result.status ~= 0 then
		-- business/UX decision:
		-- only add full tree when current status isn't OK
		-- (it's not interesting to see a detailed explanation
		-- of why everything is a-OK)
		parsed_data.data = parse_report_data(deepcopy(res))
		if parsed_data.data == false then
			SYSError("Trying to process report data for an result after executing a rule. Skipping "..obj.name)
			return false
		end
	end
	parsed_data["meta"] = {
		timestamp = timestamp,
		host = obj.materialized.host,
		name = obj.name,
		state = result.status
	}
	if type(write_callback) == "function" then
		return write_callback(parsed_data)
	elseif write_callback then
		SYSError("Expected a function as write_callback, got "..type(write_callback))
	end
	return false
end

function write_report_data_to_php_controller(parsed_data)
	-- TODO: Implement popen2 to be able to read stderr on failure.
	local handle = io.popen("/usr/bin/php /opt/monitor/op5/ninja/index.php synergy_report/insert_report_data", "w")
	if not handle then
		SYSError("Could not open file handle to synergy_report")
		return false
	end
	handle:write(json.encode(parsed_data))
	handle:close()
	return true
end

----------------------------------------------
-- Materializes not materialized service using nacoma API
-- @param host string
-- @param service string
-- @return mixed
----------------------------------------------
function materialize_service(host, service)
	if not config.daemon_materialize then
		SYSError("Wanted to materialize service ("..host..";"..service.."), but blocked by daemon config.")
		return false
	end
	local url = "/monitor/index.php/synergy/view#/tree:"
	SYSDebug("Materializing service: "..host..";"..service)
	local config_submit_output, materialized = exec_command("php "..config.nacoma_apiurl.." -t service -o template=bizproc_service -o host_name=\""..host.."\" -o service_description=\""..service.."\" -o action_url=\""..url..service.."\" -u synergy_daemon")
	if materialized ~= 0 then
		SYSError("Failed to create service: "..host..";"..service.." Output: "..config_submit_output)
		return false
	end
	local config_save_output, save = exec_command("php "..config.nacoma_apiurl.." -a save_config -u synergy_daemon")
	if save ~= 0 then
		SYSError("Failed to save config when creating service: "..host..";"..service.." Output: "..config_save_output)
		return false
	end
	return true
end

----------------------------------------------
-- Packs all arguments into table
-- @param ... mixed
-- @return table
----------------------------------------------
function pack_arguments(...)
	local r = {}
	for k,v in ipairs(arg) do
		if k ~= "n" and v then
			table.insert(r, v)
		end
	end
	return r
end

----------------------------------------------
-- Get contacts from contactgroups
-- @param groups table
-- @return table
----------------------------------------------
function contactgroups_get_contacts(groups)
	local res = {}
	if not groups then
		groups = "*"
	end
	if groups == "*" then
	return "*"
	end
	for i,name in ipairs(groups) do
		if not contactgroups_cache[name] then
			contactgroups_cache[name] = response_to_obj(mklive:request(Request.new("contactgroups"):where("name = ?", name))).members
		end
		res = table_merge(res, contactgroups_cache[name])
	end
	return res
end

----------------------------------------------
-- Merge tables
-- @param dst table
-- @param src mixed (string|array)
-- @return table
----------------------------------------------
table_merge = function(dst, src)
	if src == "*" then
		if dst == "*" then
			return "*"
		end
		table.insert(dst, "*")
		return dst
	end
	for i,v in ipairs(src) do
		table.insert(dst, v)
	end
	return dst
end
